/////////
// iOS template / framework
// Copyright 2011 Jacob Schwartz
// kevlarx@gmail.com
/////////

// NOTE: This file is based heavily off of the Mersenne, the following is the original copyright notice
/* 
 A C-program for MT19937, with initialization improved 2002/1/26.
 Coded by Takuji Nishimura and Makoto Matsumoto.
 
 Before using, initialize the state by using init_genrand(seed)  
 or init_by_array(init_key, key_length).
 
 Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
 All rights reserved.                          
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 
 3. The names of its contributors may not be used to endorse or promote 
 products derived from this software without specific prior written 
 permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 
 Any feedback is very welcome.
 http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
 email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
 */

#include <mach/mach_time.h>
#include "Random.h"

#define RANDOM_M    397
#define MATRIX_A    0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK  0x80000000UL /* most significant w-r bits */
#define LOWER_MASK  0x7fffffffUL /* least significant r bits */

Random::Random(uint32_t s/*=0*/) {
    if (s == 0) {
        seed( mach_absolute_time() ); // using a faster timer than time() in cause someone seeds two RNGs in the same second
    } else {
        seed(s);
    }
}

Random::~Random() {
}

void Random::seed(uint32_t s) {
    m_mt[0]= s & 0xffffffffUL;
    for (m_mti=1; m_mti<RANDOM_N; m_mti++) {
        m_mt[m_mti] = 
	    (1812433253UL * (m_mt[m_mti-1] ^ (m_mt[m_mti-1] >> 30)) + m_mti); 
        /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
        /* In the previous versions, MSBs of the seed affect   */
        /* only MSBs of the array mt[].                        */
        /* 2002/01/09 modified by Makoto Matsumoto             */
        m_mt[m_mti] &= 0xffffffffUL;
        /* for >32 bit machines */
    }
}

/* generates a random number on [0,0xffffffff]-interval */
uint32_t Random::rand() {
    unsigned long y;
    static unsigned long mag01[2]={0x0UL, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */
    
    if (m_mti >= RANDOM_N) { /* generate N words at one time */
        int kk;
        
        if (m_mti == RANDOM_N+1)   /* if init_genrand() has not been called, */
            seed(5489UL); /* a default initial seed is used */
        
        for (kk=0;kk<RANDOM_N-RANDOM_M;kk++) {
            y = (m_mt[kk]&UPPER_MASK)|(m_mt[kk+1]&LOWER_MASK);
            m_mt[kk] = m_mt[kk+RANDOM_M] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        for (;kk<RANDOM_N-1;kk++) {
            y = (m_mt[kk]&UPPER_MASK)|(m_mt[kk+1]&LOWER_MASK);
            m_mt[kk] = m_mt[kk+(RANDOM_M-RANDOM_N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        y = (m_mt[RANDOM_N-1]&UPPER_MASK)|(m_mt[0]&LOWER_MASK);
        m_mt[RANDOM_N-1] = m_mt[RANDOM_M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];
        
        m_mti = 0;
    }
    
    y = m_mt[m_mti++];
    
    /* Tempering */
    y ^= (y >> 11);
    y ^= (y << 7) & 0x9d2c5680UL;
    y ^= (y << 15) & 0xefc60000UL;
    y ^= (y >> 18);
    
    return y;
}

// get number in [low, high)
int32_t Random::randRange(int32_t low, int32_t high) {
//    dbgAssert(low <= high);
    return (high-low)*randFloat() + low;
}

int32_t Random::randInt(int32_t high) {
//    dbgAssert(high > 0);
    return high*randFloat();
}

/* generates a random number on [0,1)-real-interval */
float Random::randFloat() {
    return rand()*(1.0/4294967296.0); 
    /* divided by 2^32 */
}
